/*
 * livepatch_bsc1204381
 *
 * Fix for bsc#1204381
 *
 *  Upstream commit:
 *  dcd46d897adb ("exec: Force single empty string when argv is empty")
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commit:
 *  4ee3bddf6bfe70d01f64ba87c08ed67a312b92f1
 *
 *  SLE15-SP2 and -SP3 commit:
 *  dffa04e1f4ac951749521cdd317ea4def7e13599
 *
 *  SLE15-SP4 commit:
 *  256509dd34346035da0e108eb0e3f004752fc224
 *
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* klp-ccp: from fs/exec.c */
#include <linux/kernel_read_file.h>
#include <linux/slab.h>
#include <linux/kernel.h>
#include <linux/file.h>
#include <linux/fdtable.h>
#include <linux/sched.h>

/* klp-ccp: from include/linux/fs.h */
static struct filename *(*klpe_getname_flags)(const char __user *, int, int *);
static struct filename *(*klpe_getname)(const char __user *);
static struct filename *(*klpe_getname_kernel)(const char *);
static void (*klpe_putname)(struct filename *name);

/* klp-ccp: from fs/exec.c */
#include <linux/mm.h>
#include <linux/stat.h>
#include <linux/fcntl.h>
#include <linux/swap.h>

/* klp-ccp: from include/linux/user_namespace.h */
static bool (*klpe_is_ucounts_overlimit)(struct ucounts *ucounts, enum ucount_type type, unsigned long max);

/* klp-ccp: from fs/exec.c */
#include <linux/highmem.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/sched/mm.h>
#include <linux/sched/coredump.h>
#include <linux/sched/signal.h>
#include <linux/sched/task.h>
#include <linux/pagemap.h>
#include <linux/perf_event.h>
#include <linux/highmem.h>
#include <linux/spinlock.h>
#include <linux/key.h>
#include <linux/personality.h>
#include <linux/binfmts.h>

/* klp-ccp: from include/linux/binfmts.h */
int klpp_kernel_execve(const char *filename,
		  const char *const *argv, const char *const *envp);

/* klp-ccp: from fs/exec.c */
#include <linux/pid_namespace.h>
#include <linux/module.h>
#include <linux/namei.h>
#include <linux/mount.h>
#include <linux/security.h>
#include <linux/kmod.h>
#include <linux/syscalls.h>
#include <linux/compat.h>
#include <linux/vmalloc.h>
#include <linux/syscall_user_dispatch.h>
#include <linux/uaccess.h>

#ifdef CONFIG_MMU

static struct page *(*klpe_get_arg_page)(struct linux_binprm *bprm, unsigned long pos,
		int write);

static void put_arg_page(struct page *page)
{
	put_page(page);
}

static void flush_arg_page(struct linux_binprm *bprm, unsigned long pos,
		struct page *page)
{
	flush_cache_page(bprm->vma, pos, page_to_pfn(page));
}

static bool valid_arg_len(struct linux_binprm *bprm, long len)
{
	return len <= MAX_ARG_STRLEN;
}

#else
#error "klp-ccp: non-taken branch"
#endif /* CONFIG_MMU */

struct user_arg_ptr {
#ifdef CONFIG_COMPAT
	bool is_compat;
#endif
	union {
		const char __user *const __user *native;
#ifdef CONFIG_COMPAT
		const compat_uptr_t __user *compat;
#endif
	} ptr;
};

static const char __user *get_user_arg_ptr(struct user_arg_ptr argv, int nr)
{
	const char __user *native;

#ifdef CONFIG_COMPAT
	if (unlikely(argv.is_compat)) {
		compat_uptr_t compat;

		if (get_user(compat, argv.ptr.compat + nr))
			return ERR_PTR(-EFAULT);

		return compat_ptr(compat);
	}
#endif
	if (get_user(native, argv.ptr.native + nr))
		return ERR_PTR(-EFAULT);

	return native;
}

static int count(struct user_arg_ptr argv, int max)
{
	int i = 0;

	if (argv.ptr.native != NULL) {
		for (;;) {
			const char __user *p = get_user_arg_ptr(argv, i);

			if (!p)
				break;

			if (IS_ERR(p))
				return -EFAULT;

			if (i >= max)
				return -E2BIG;
			++i;

			if (fatal_signal_pending(current))
				return -ERESTARTNOHAND;
			cond_resched();
		}
	}
	return i;
}

static int count_strings_kernel(const char *const *argv)
{
	int i;

	if (!argv)
		return 0;

	for (i = 0; argv[i]; ++i) {
		if (i >= MAX_ARG_STRINGS)
			return -E2BIG;
		if (fatal_signal_pending(current))
			return -ERESTARTNOHAND;
		cond_resched();
	}
	return i;
}

static int klpp_bprm_stack_limits(struct linux_binprm *bprm)
{
	unsigned long limit, ptr_size;

	/*
	 * Limit to 1/4 of the max stack size or 3/4 of _STK_LIM
	 * (whichever is smaller) for the argv+env strings.
	 * This ensures that:
	 *  - the remaining binfmt code will not run out of stack space,
	 *  - the program will have a reasonable amount of stack left
	 *    to work from.
	 */
	limit = _STK_LIM / 4 * 3;
	limit = min(limit, bprm->rlim_stack.rlim_cur / 4);
	/*
	 * We've historically supported up to 32 pages (ARG_MAX)
	 * of argument strings even with small stacks
	 */
	limit = max_t(unsigned long, limit, ARG_MAX);
	/*
	 * We must account for the size of all the argv and envp pointers to
	 * the argv and envp strings, since they will also take up space in
	 * the stack. They aren't stored until much later when we can't
	 * signal to the parent that the child has run out of stack space.
	 * Instead, calculate it here so it's possible to fail gracefully.
	 */
	/*
	 * Fix bsc#1204381
	 *  -1 line, +1 line
	 */
	ptr_size = (max(bprm->argc, 1) + bprm->envc) * sizeof(void *);
	if (limit <= ptr_size)
		return -E2BIG;
	limit -= ptr_size;

	bprm->argmin = bprm->p - limit;
	return 0;
}

static int klpr_copy_strings(int argc, struct user_arg_ptr argv,
			struct linux_binprm *bprm)
{
	struct page *kmapped_page = NULL;
	char *kaddr = NULL;
	unsigned long kpos = 0;
	int ret;

	while (argc-- > 0) {
		const char __user *str;
		int len;
		unsigned long pos;

		ret = -EFAULT;
		str = get_user_arg_ptr(argv, argc);
		if (IS_ERR(str))
			goto out;

		len = strnlen_user(str, MAX_ARG_STRLEN);
		if (!len)
			goto out;

		ret = -E2BIG;
		if (!valid_arg_len(bprm, len))
			goto out;

		/* We're going to work our way backwords. */
		pos = bprm->p;
		str += len;
		bprm->p -= len;
#ifdef CONFIG_MMU
		if (bprm->p < bprm->argmin)
			goto out;
#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
		while (len > 0) {
			int offset, bytes_to_copy;

			if (fatal_signal_pending(current)) {
				ret = -ERESTARTNOHAND;
				goto out;
			}
			cond_resched();

			offset = pos % PAGE_SIZE;
			if (offset == 0)
				offset = PAGE_SIZE;

			bytes_to_copy = offset;
			if (bytes_to_copy > len)
				bytes_to_copy = len;

			offset -= bytes_to_copy;
			pos -= bytes_to_copy;
			str -= bytes_to_copy;
			len -= bytes_to_copy;

			if (!kmapped_page || kpos != (pos & PAGE_MASK)) {
				struct page *page;

				page = (*klpe_get_arg_page)(bprm, pos, 1);
				if (!page) {
					ret = -E2BIG;
					goto out;
				}

				if (kmapped_page) {
					flush_dcache_page(kmapped_page);
					kunmap(kmapped_page);
					put_arg_page(kmapped_page);
				}
				kmapped_page = page;
				kaddr = kmap(kmapped_page);
				kpos = pos & PAGE_MASK;
				flush_arg_page(bprm, kpos, kmapped_page);
			}
			if (copy_from_user(kaddr+offset, str, bytes_to_copy)) {
				ret = -EFAULT;
				goto out;
			}
		}
	}
	ret = 0;
out:
	if (kmapped_page) {
		flush_dcache_page(kmapped_page);
		kunmap(kmapped_page);
		put_arg_page(kmapped_page);
	}
	return ret;
}

static int (*klpe_copy_strings_kernel)(int argc, const char *const *argv,
			       struct linux_binprm *bprm);

static void (*klpe_free_bprm)(struct linux_binprm *bprm);

static struct linux_binprm *(*klpe_alloc_bprm)(int fd, struct filename *filename);

static int (*klpe_bprm_execve)(struct linux_binprm *bprm,
		       int fd, struct filename *filename, int flags);

static int klpp_do_execveat_common(int fd, struct filename *filename,
			      struct user_arg_ptr argv,
			      struct user_arg_ptr envp,
			      int flags)
{
	struct linux_binprm *bprm;
	int retval;

	if (IS_ERR(filename))
		return PTR_ERR(filename);

	/*
	 * We move the actual failure in case of RLIMIT_NPROC excess from
	 * set*uid() to execve() because too many poorly written programs
	 * don't check setuid() return code.  Here we additionally recheck
	 * whether NPROC limit is still exceeded.
	 */
	if ((current->flags & PF_NPROC_EXCEEDED) &&
	    (*klpe_is_ucounts_overlimit)(current_ucounts(), UCOUNT_RLIMIT_NPROC, rlimit(RLIMIT_NPROC))) {
		retval = -EAGAIN;
		goto out_ret;
	}

	/* We're below the limit (still or again), so we don't want to make
	 * further execve() calls fail. */
	current->flags &= ~PF_NPROC_EXCEEDED;

	bprm = (*klpe_alloc_bprm)(fd, filename);
	if (IS_ERR(bprm)) {
		retval = PTR_ERR(bprm);
		goto out_ret;
	}

	retval = count(argv, MAX_ARG_STRINGS);
	/*
	 * Fix bsc#1204381
	 *  +3 lines
	 */
	if (retval == 0)
		pr_warn_once("process '%s' launched '%s' with NULL argv: empty string added\n",
			     current->comm, bprm->filename);

	if (retval < 0)
		goto out_free;
	bprm->argc = retval;

	retval = count(envp, MAX_ARG_STRINGS);
	if (retval < 0)
		goto out_free;
	bprm->envc = retval;

	retval = klpp_bprm_stack_limits(bprm);
	if (retval < 0)
		goto out_free;

	retval = copy_string_kernel(bprm->filename, bprm);
	if (retval < 0)
		goto out_free;
	bprm->exec = bprm->p;

	retval = klpr_copy_strings(bprm->envc, envp, bprm);
	if (retval < 0)
		goto out_free;

	retval = klpr_copy_strings(bprm->argc, argv, bprm);
	if (retval < 0)
		goto out_free;

	/*
	 * Fix bsc#1204381
	 *  +13 lines
	 */
	/*
	 * When argv is empty, add an empty string ("") as argv[0] to
	 * ensure confused userspace programs that start processing
	 * from argv[1] won't end up walking envp. See also
	 * bprm_stack_limits().
	 */
	if (bprm->argc == 0) {
		retval = copy_string_kernel("", bprm);
		if (retval < 0)
			goto out_free;
		bprm->argc = 1;
	}

	retval = (*klpe_bprm_execve)(bprm, fd, filename, flags);
out_free:
	(*klpe_free_bprm)(bprm);

out_ret:
	(*klpe_putname)(filename);
	return retval;
}

int klpp_kernel_execve(const char *kernel_filename,
		  const char *const *argv, const char *const *envp)
{
	struct filename *filename;
	struct linux_binprm *bprm;
	int fd = AT_FDCWD;
	int retval;

	filename = (*klpe_getname_kernel)(kernel_filename);
	if (IS_ERR(filename))
		return PTR_ERR(filename);

	bprm = (*klpe_alloc_bprm)(fd, filename);
	if (IS_ERR(bprm)) {
		retval = PTR_ERR(bprm);
		goto out_ret;
	}

	retval = count_strings_kernel(argv);
	if (retval < 0)
		goto out_free;
	bprm->argc = retval;

	retval = count_strings_kernel(envp);
	if (retval < 0)
		goto out_free;
	bprm->envc = retval;

	retval = klpp_bprm_stack_limits(bprm);
	if (retval < 0)
		goto out_free;

	retval = copy_string_kernel(bprm->filename, bprm);
	if (retval < 0)
		goto out_free;
	bprm->exec = bprm->p;

	retval = (*klpe_copy_strings_kernel)(bprm->envc, envp, bprm);
	if (retval < 0)
		goto out_free;

	retval = (*klpe_copy_strings_kernel)(bprm->argc, argv, bprm);
	if (retval < 0)
		goto out_free;

	retval = (*klpe_bprm_execve)(bprm, fd, filename, 0);
out_free:
	(*klpe_free_bprm)(bprm);
out_ret:
	(*klpe_putname)(filename);
	return retval;
}

static int klpp_do_execve(struct filename *filename,
	const char __user *const __user *__argv,
	const char __user *const __user *__envp)
{
	struct user_arg_ptr argv = { .ptr.native = __argv };
	struct user_arg_ptr envp = { .ptr.native = __envp };
	return klpp_do_execveat_common(AT_FDCWD, filename, argv, envp, 0);
}

static int klpp_do_execveat(int fd, struct filename *filename,
		const char __user *const __user *__argv,
		const char __user *const __user *__envp,
		int flags)
{
	struct user_arg_ptr argv = { .ptr.native = __argv };
	struct user_arg_ptr envp = { .ptr.native = __envp };

	return klpp_do_execveat_common(fd, filename, argv, envp, flags);
}

#ifdef CONFIG_COMPAT
static int klpp_compat_do_execve(struct filename *filename,
	const compat_uptr_t __user *__argv,
	const compat_uptr_t __user *__envp)
{
	struct user_arg_ptr argv = {
		.is_compat = true,
		.ptr.compat = __argv,
	};
	struct user_arg_ptr envp = {
		.is_compat = true,
		.ptr.compat = __envp,
	};
	return klpp_do_execveat_common(AT_FDCWD, filename, argv, envp, 0);
}

static int klpp_compat_do_execveat(int fd, struct filename *filename,
			      const compat_uptr_t __user *__argv,
			      const compat_uptr_t __user *__envp,
			      int flags)
{
	struct user_arg_ptr argv = {
		.is_compat = true,
		.ptr.compat = __argv,
	};
	struct user_arg_ptr envp = {
		.is_compat = true,
		.ptr.compat = __envp,
	};
	return klpp_do_execveat_common(fd, filename, argv, envp, flags);
}
#endif

__SYSCALL_DEFINEx(3, _klpp_execve,
		const char __user *, filename,
		const char __user *const __user *, argv,
		const char __user *const __user *, envp)
{
	return klpp_do_execve((*klpe_getname)(filename), argv, envp);
}

__SYSCALL_DEFINEx(5, _klpp_execveat,
		int, fd, const char __user *, filename,
		const char __user *const __user *, argv,
		const char __user *const __user *, envp,
		int, flags)
{
	int lookup_flags = (flags & AT_EMPTY_PATH) ? LOOKUP_EMPTY : 0;

	return klpp_do_execveat(fd,
			   (*klpe_getname_flags)(filename, lookup_flags, NULL),
			   argv, envp, flags);
}

#ifdef CONFIG_COMPAT

COMPAT_SYSCALL_DEFINEx(3, _klpp_execve, const char __user *, filename,
	const compat_uptr_t __user *, argv,
	const compat_uptr_t __user *, envp)
{
	return klpp_compat_do_execve((*klpe_getname)(filename), argv, envp);
}

COMPAT_SYSCALL_DEFINEx(5, _klpp_execveat, int, fd,
		       const char __user *, filename,
		       const compat_uptr_t __user *, argv,
		       const compat_uptr_t __user *, envp,
		       int,  flags)
{
	int lookup_flags = (flags & AT_EMPTY_PATH) ? LOOKUP_EMPTY : 0;

	return klpp_compat_do_execveat(fd,
				  (*klpe_getname_flags)(filename, lookup_flags, NULL),
				  argv, envp, flags);
}

#endif



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1204381.h"
#include "../kallsyms_relocs.h"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "alloc_bprm", (void *)&klpe_alloc_bprm },
	{ "bprm_execve", (void *)&klpe_bprm_execve },
	{ "copy_strings_kernel", (void *)&klpe_copy_strings_kernel },
	{ "free_bprm", (void *)&klpe_free_bprm },
	{ "get_arg_page", (void *)&klpe_get_arg_page },
	{ "getname", (void *)&klpe_getname },
	{ "getname_flags", (void *)&klpe_getname_flags },
	{ "getname_kernel", (void *)&klpe_getname_kernel },
	{ "is_ucounts_overlimit", (void *)&klpe_is_ucounts_overlimit },
	{ "putname", (void *)&klpe_putname },
};

int livepatch_bsc1204381_init(void)
{
	return klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
}
